from __future__ import annotations

import numpy as np
import pytest

from contourpy import FillType, LineType
from contourpy.typecheck import (
    check_code_array, check_filled, check_lines, check_offset_array, check_point_array,
)
from contourpy.types import code_dtype, offset_dtype, point_dtype


def test_check_code_array() -> None:
    # Valid code array, does not raise.
    check_code_array(np.array([1, 2, 79], dtype=code_dtype))

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        check_code_array([1, 2, 79])

    with pytest.raises(ValueError, match=r"Expected numpy array of dtype <class 'numpy.uint8'>"):
        check_code_array(np.array([1, 2, 79], dtype=np.int64))

    with pytest.raises(ValueError, match=r"Expected numpy array of shape"):
        check_code_array(np.array([[1, 2, 79]], dtype=code_dtype))

    with pytest.raises(ValueError, match=r"First element of code array must be 1"):
        check_code_array(np.array([2, 2, 79], dtype=code_dtype))


def test_check_offset_array() -> None:
    # Valid offset array, does not raise.
    check_offset_array(np.array([0, 5], dtype=offset_dtype))

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        check_offset_array([0, 5])

    with pytest.raises(ValueError, match=r"Expected numpy array of dtype <class 'numpy.uint32'>"):
        check_offset_array(np.array([0, 5], dtype=np.int64))

    with pytest.raises(ValueError, match=r"Expected numpy array of shape"):
        check_offset_array(np.array([[0], [5]], dtype=offset_dtype))

    with pytest.raises(ValueError, match=r"First element of offset array must be 0"):
        check_offset_array(np.array([1, 5], dtype=offset_dtype))


def test_check_point_array() -> None:
    # Valid point array, does not raise.
    check_point_array(np.array([[1.1, 2.2], [3.3, 4.4]], dtype=point_dtype))

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        check_point_array([[1.1, 2.2], [3.3, 4.4]])

    with pytest.raises(ValueError, match=r"Expected numpy array of dtype <class 'numpy.float64'>"):
        check_point_array(np.array([[1.1, 2.2], [3.3, 4.4]], dtype=np.float32))

    with pytest.raises(ValueError, match=r"Expected numpy array of shape"):
        check_point_array(np.array([[1.1, 2.2, 3.3, 4.4]], dtype=point_dtype))


def test_check_filled_OuterCode() -> None:
    # Valid filled, does not raise.
    points = np.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6]], dtype=point_dtype)
    codes = np.array([1, 2, 79], dtype=code_dtype)
    check_filled(([points], [codes]), FillType.OuterCode)

    with pytest.raises(TypeError, match=r"Expected tuple not <class 'list'>"):
        check_filled(list(), FillType.OuterCode)
    with pytest.raises(ValueError, match=r"Expected tuple of length 2 not 3"):
        check_filled(([], [], []), FillType.OuterCode)
    with pytest.raises(TypeError, match=r"Expected tuple to contain 2 lists"):
        check_filled(([], None), FillType.OuterCode)
    with pytest.raises(ValueError, match=r"Expected 2 lists with same length"):
        check_filled(([1, 2, 3], [1, 2]), FillType.OuterCode)

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_filled(([list(points)], [codes]), FillType.OuterCode)
    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_filled(([points], [list(codes)]), FillType.OuterCode)

    with pytest.raises(ValueError, match=r"Points and codes have different lengths in polygon 0"):
        check_filled(([points], [codes[:-1]]), FillType.OuterCode)


def test_check_filled_OuterOffset() -> None:
    # Valid filled, does not raise.
    points = np.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6], [7.7, 8.8]], dtype=point_dtype)
    offsets = np.array([0, 2, 4], dtype=offset_dtype)
    check_filled(([points], [offsets]), FillType.OuterOffset)

    with pytest.raises(TypeError, match=r"Expected tuple not <class 'list'>"):
        check_filled(list(), FillType.OuterOffset)
    with pytest.raises(ValueError, match=r"Expected tuple of length 2 not 3"):
        check_filled(([], [], []), FillType.OuterOffset)
    with pytest.raises(TypeError, match=r"Expected tuple to contain 2 lists"):
        check_filled(([], None), FillType.OuterOffset)
    with pytest.raises(ValueError, match=r"Expected 2 lists with same length"):
        check_filled(([1, 2, 3], [1, 2]), FillType.OuterOffset)

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_filled(([list(points)], [offsets]), FillType.OuterOffset)
    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_filled(([points], [list(offsets)]), FillType.OuterOffset)

    with pytest.raises(ValueError, match=r"Inconsistent points and offsets in polygon 0"):
        check_filled(([points], [offsets[:-1]]), FillType.OuterOffset)


def test_check_filled_ChunkCombinedCode() -> None:
    # Valid filled, does not raise.
    points = np.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6]], dtype=point_dtype)
    codes = np.array([1, 2, 79], dtype=code_dtype)
    check_filled(([points], [codes]), FillType.ChunkCombinedCode)
    check_filled(([None], [None]), FillType.ChunkCombinedCode)
    check_filled(([None, points], [None, codes]), FillType.ChunkCombinedCode)

    with pytest.raises(TypeError, match=r"Expected tuple not <class 'list'>"):
        check_filled(list(), FillType.ChunkCombinedCode)
    with pytest.raises(ValueError, match=r"Expected tuple of length 2 not 3"):
        check_filled(([], [], []), FillType.ChunkCombinedCode)
    with pytest.raises(TypeError, match=r"Expected tuple to contain 2 lists"):
        check_filled(([], None), FillType.ChunkCombinedCode)
    with pytest.raises(ValueError, match=r"Expected 2 lists with same length"):
        check_filled(([1, 2, 3], [1, 2]), FillType.ChunkCombinedCode)
    with pytest.raises(ValueError, match=r"Expected 2 non-empty lists"):
        check_filled(([], []), FillType.ChunkCombinedCode)

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_filled(([list(points)], [codes]), FillType.ChunkCombinedCode)
    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_filled(([points], [list(codes)]), FillType.ChunkCombinedCode)

    with pytest.raises(ValueError, match=r"Points and codes have different lengths in chunk 1"):
        check_filled(([None, points], [None, codes[:-1]]), FillType.ChunkCombinedCode)

    with pytest.raises(ValueError, match=r"Inconsistent Nones in chunk 0"):
        check_filled(([points], [None]), FillType.ChunkCombinedCode)
    with pytest.raises(ValueError, match=r"Inconsistent Nones in chunk 0"):
        check_filled(([None], [codes]), FillType.ChunkCombinedCode)


def test_check_filled_ChunkCombinedOffset() -> None:
    # Valid filled, does not raise.
    points = np.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6], [7.7, 8.8]], dtype=point_dtype)
    offsets = np.array([0, 2, 4], dtype=offset_dtype)
    check_filled(([points], [offsets]), FillType.ChunkCombinedOffset)
    check_filled(([None], [None]), FillType.ChunkCombinedOffset)
    check_filled(([None, points], [None, offsets]), FillType.ChunkCombinedOffset)

    with pytest.raises(TypeError, match=r"Expected tuple not <class 'list'>"):
        check_filled(list(), FillType.ChunkCombinedOffset)
    with pytest.raises(ValueError, match=r"Expected tuple of length 2 not 3"):
        check_filled(([], [], []), FillType.ChunkCombinedOffset)
    with pytest.raises(TypeError, match=r"Expected tuple to contain 2 lists"):
        check_filled(([], None), FillType.ChunkCombinedOffset)
    with pytest.raises(ValueError, match=r"Expected 2 lists with same length"):
        check_filled(([1, 2, 3], [1, 2]), FillType.ChunkCombinedOffset)
    with pytest.raises(ValueError, match=r"Expected 2 non-empty lists"):
        check_filled(([], []), FillType.ChunkCombinedOffset)

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_filled(([list(points)], [offsets]), FillType.ChunkCombinedOffset)
    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_filled(([points], [list(offsets)]), FillType.ChunkCombinedOffset)

    with pytest.raises(ValueError, match=r"Inconsistent points and offsets in chunk 1"):
        check_filled(([None, points], [None, offsets[:-1]]), FillType.ChunkCombinedOffset)

    with pytest.raises(ValueError, match=r"Inconsistent Nones in chunk 0"):
        check_filled(([points], [None]), FillType.ChunkCombinedOffset)
    with pytest.raises(ValueError, match=r"Inconsistent Nones in chunk 0"):
        check_filled(([None], [offsets]), FillType.ChunkCombinedOffset)


def test_check_filled_ChunkCombinedCodeOffset() -> None:
    # Valid filled, does not raise.
    points = np.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6], [7.7, 8.8]], dtype=point_dtype)
    codes = np.array([1, 2, 1, 79], dtype=code_dtype)
    outer_offsets = np.array([0, 4], dtype=offset_dtype)
    check_filled(([points], [codes], [outer_offsets]), FillType.ChunkCombinedCodeOffset)
    check_filled(([None], [None], [None]), FillType.ChunkCombinedCodeOffset)
    check_filled(([None, points], [None, codes], [None, outer_offsets]),
                 FillType.ChunkCombinedCodeOffset)

    with pytest.raises(TypeError, match=r"Expected tuple not <class 'list'>"):
        check_filled(list(), FillType.ChunkCombinedCodeOffset)
    with pytest.raises(ValueError, match=r"Expected tuple of length 3 not 2"):
        check_filled(([], []), FillType.ChunkCombinedCodeOffset)
    with pytest.raises(TypeError, match=r"Expected tuple to contain 3 lists"):
        check_filled(([], None, []), FillType.ChunkCombinedCodeOffset)
    with pytest.raises(ValueError, match=r"Expected 3 lists with same length"):
        check_filled(([1, 2, 3], [1, 2], [1, 2, 3]), FillType.ChunkCombinedCodeOffset)
    with pytest.raises(ValueError, match=r"Expected 3 non-empty lists"):
        check_filled(([], [], []), FillType.ChunkCombinedCodeOffset)

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_filled(([list(points)], [codes], [outer_offsets]), FillType.ChunkCombinedCodeOffset)
    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_filled(([points], [list(codes)], [outer_offsets]), FillType.ChunkCombinedCodeOffset)
    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_filled(([points], [codes], [list(outer_offsets)]), FillType.ChunkCombinedCodeOffset)

    with pytest.raises(ValueError, match=r"Points and codes have different lengths in chunk 1"):
        check_filled(([None, points], [None, codes[:-1]], [None, outer_offsets]),
                     FillType.ChunkCombinedCodeOffset)
    with pytest.raises(ValueError, match=r"Inconsistent codes and outer_offsets in chunk 0"):
        check_filled(([points], [codes], [np.array([0, 4, 5], dtype=offset_dtype)]),
                     FillType.ChunkCombinedCodeOffset)

    with pytest.raises(ValueError, match=r"Inconsistent Nones in chunk 0"):
        check_filled(([points], [codes], [None]), FillType.ChunkCombinedCodeOffset)
    with pytest.raises(ValueError, match=r"Inconsistent Nones in chunk 0"):
        check_filled(([points], [None], [outer_offsets]), FillType.ChunkCombinedCodeOffset)
    with pytest.raises(ValueError, match=r"Inconsistent Nones in chunk 0"):
        check_filled(([None], [codes], [outer_offsets]), FillType.ChunkCombinedCodeOffset)


def test_check_filled_ChunkCombinedOffsetOffset() -> None:
    # Valid filled, does not raise.
    points = np.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6], [7.7, 8.8]], dtype=point_dtype)
    offsets = np.array([0, 2, 4], dtype=offset_dtype)
    outer_offsets = np.array([0, 2], dtype=offset_dtype)
    check_filled(([points], [offsets], [outer_offsets]), FillType.ChunkCombinedOffsetOffset)
    check_filled(([None], [None], [None]), FillType.ChunkCombinedOffsetOffset)
    check_filled(([None, points], [None, offsets], [None, outer_offsets]),
                 FillType.ChunkCombinedOffsetOffset)

    with pytest.raises(TypeError, match=r"Expected tuple not <class 'list'>"):
        check_filled(list(), FillType.ChunkCombinedOffsetOffset)
    with pytest.raises(ValueError, match=r"Expected tuple of length 3 not 2"):
        check_filled(([], []), FillType.ChunkCombinedOffsetOffset)
    with pytest.raises(TypeError, match=r"Expected tuple to contain 3 lists"):
        check_filled(([], None, []), FillType.ChunkCombinedOffsetOffset)
    with pytest.raises(ValueError, match=r"Expected 3 lists with same length"):
        check_filled(([1, 2, 3], [1, 2], [1, 2, 3]), FillType.ChunkCombinedOffsetOffset)
    with pytest.raises(ValueError, match=r"Expected 3 non-empty lists"):
        check_filled(([], [], []), FillType.ChunkCombinedOffsetOffset)

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_filled(([list(points)], [offsets], [outer_offsets]),
                      FillType.ChunkCombinedOffsetOffset)
    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_filled(([points], [list(offsets)], [outer_offsets]),
                      FillType.ChunkCombinedOffsetOffset)
    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_filled(([points], [offsets], [list(outer_offsets)]),
                      FillType.ChunkCombinedOffsetOffset)

    with pytest.raises(ValueError, match=r"Inconsistent points and offsets in chunk 1"):
        check_filled(([None, points], [None, offsets[:-1]], [None, outer_offsets]),
                     FillType.ChunkCombinedOffsetOffset)
    with pytest.raises(ValueError, match=r"Inconsistent offsets and outer_offsets in chunk 0"):
        check_filled(([points], [offsets], [np.array([0, 4, 5], dtype=offset_dtype)]),
                     FillType.ChunkCombinedOffsetOffset)

    with pytest.raises(ValueError, match=r"Inconsistent Nones in chunk 0"):
        check_filled(([points], [offsets], [None]), FillType.ChunkCombinedOffsetOffset)
    with pytest.raises(ValueError, match=r"Inconsistent Nones in chunk 0"):
        check_filled(([points], [None], [outer_offsets]), FillType.ChunkCombinedOffsetOffset)
    with pytest.raises(ValueError, match=r"Inconsistent Nones in chunk 0"):
        check_filled(([None], [offsets], [outer_offsets]), FillType.ChunkCombinedOffsetOffset)


def test_check_lines_Separate() -> None:
    # Valid lines, does not raise.
    points = np.array([[1.1, 2.2], [3.3, 4.4]], dtype=point_dtype)
    check_lines([points], LineType.Separate)

    with pytest.raises(TypeError, match=r"Expected list not <class 'tuple'>"):
        check_lines(tuple(), LineType.Separate)

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        check_lines([[]], LineType.Separate)


def test_check_lines_SeparateCode() -> None:
    # Valid lines, does not raise.
    points = np.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6]], dtype=point_dtype)
    codes = np.array([1, 2, 79], dtype=code_dtype)
    check_lines(([points], [codes]), LineType.SeparateCode)

    with pytest.raises(TypeError, match=r"Expected tuple not <class 'list'>"):
        check_lines(list(), LineType.SeparateCode)
    with pytest.raises(ValueError, match=r"Expected tuple of length 2 not 3"):
        check_lines(([], [], []), LineType.SeparateCode)
    with pytest.raises(TypeError, match=r"Expected tuple to contain 2 lists"):
        check_lines(([], None), LineType.SeparateCode)
    with pytest.raises(ValueError, match=r"Expected 2 lists with same length"):
        check_lines(([1, 2, 3], [1, 2]), LineType.SeparateCode)

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_lines(([list(points)], [codes]), LineType.SeparateCode)
    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'"):
        check_lines(([points], [list(codes)]), LineType.SeparateCode)

    with pytest.raises(ValueError, match=r"Points and codes have different lengths in line 0"):
        check_lines(([points], [codes[:-1]]), LineType.SeparateCode)


def test_check_lines_ChunkCombinedCode() -> None:
    # Valid lines, does not raise.
    points = np.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6], [7.7, 8.8]], dtype=point_dtype)
    codes = np.array([1, 2, 1, 2], dtype=code_dtype)
    check_lines(([points], [codes]), LineType.ChunkCombinedCode)
    check_lines(([None], [None]), LineType.ChunkCombinedCode)
    check_lines(([None, points], [None, codes]), LineType.ChunkCombinedCode)

    with pytest.raises(TypeError, match=r"Expected tuple not <class 'list'>"):
        check_lines(list(), LineType.ChunkCombinedCode)
    with pytest.raises(ValueError, match=r"Expected tuple of length 2 not 3"):
        check_lines(([], [], []), LineType.ChunkCombinedCode)
    with pytest.raises(TypeError, match=r"Expected tuple to contain 2 lists"):
        check_lines(([], None), LineType.ChunkCombinedCode)
    with pytest.raises(ValueError, match=r"Expected 2 lists with same length"):
        check_lines(([None, None], [None]), LineType.ChunkCombinedCode)
    with pytest.raises(ValueError, match=r"Expected 2 non-empty lists"):
        check_lines(([], []), LineType.ChunkCombinedCode)

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        check_lines(([list(points)], [codes]), LineType.ChunkCombinedCode)
    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        check_lines(([points], [list(codes)]), LineType.ChunkCombinedCode)

    with pytest.raises(ValueError, match=r"Points and codes have different lengths in chunk 1"):
        check_lines(([None, points], [None, codes[:-1]]), LineType.ChunkCombinedCode)

    with pytest.raises(ValueError, match=r"Inconsistent Nones in chunk 0"):
        check_lines(([points], [None]), LineType.ChunkCombinedCode)
    with pytest.raises(ValueError, match=r"Inconsistent Nones in chunk 0"):
        check_lines(([None], [codes]), LineType.ChunkCombinedCode)


def test_check_lines_ChunkCombinedOffset() -> None:
    # Valid lines, does not raise.
    points = np.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6], [7.7, 8.8]], dtype=point_dtype)
    offsets = np.array([0, 2, 4], dtype=offset_dtype)
    check_lines(([points], [offsets]), LineType.ChunkCombinedOffset)
    check_lines(([None], [None]), LineType.ChunkCombinedOffset)
    check_lines(([None, points], [None, offsets]), LineType.ChunkCombinedOffset)

    with pytest.raises(TypeError, match=r"Expected tuple not <class 'list'>"):
        check_lines(list(), LineType.ChunkCombinedOffset)
    with pytest.raises(ValueError, match=r"Expected tuple of length 2 not 3"):
        check_lines(([], [], []), LineType.ChunkCombinedOffset)
    with pytest.raises(TypeError, match=r"Expected tuple to contain 2 lists"):
        check_lines(([], None), LineType.ChunkCombinedOffset)
    with pytest.raises(ValueError, match=r"Expected 2 lists with same length"):
        check_lines(([None, None], [None]), LineType.ChunkCombinedOffset)
    with pytest.raises(ValueError, match=r"Expected 2 non-empty lists"):
        check_lines(([], []), LineType.ChunkCombinedOffset)

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        check_lines(([list(points)], [offsets]), LineType.ChunkCombinedOffset)
    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        check_lines(([points], [list(offsets)]), LineType.ChunkCombinedOffset)

    with pytest.raises(ValueError, match=r"Inconsistent points and offsets in chunk 1"):
        check_lines(([None, points], [None, offsets[:-1]]), LineType.ChunkCombinedOffset)

    with pytest.raises(ValueError, match=r"Inconsistent Nones in chunk 0"):
        check_lines(([points], [None]), LineType.ChunkCombinedOffset)
    with pytest.raises(ValueError, match=r"Inconsistent Nones in chunk 0"):
        check_lines(([None], [offsets]), LineType.ChunkCombinedOffset)


def test_check_lines_ChunkCombinedNan() -> None:
    # Valid lines, does not raise.
    points = np.array([[1.1, 2.2], [3.3, 4.4], [np.nan, np.nan], [5.5, 6.6], [7.7, 8.8]],
                       dtype=point_dtype)
    check_lines(([points],), LineType.ChunkCombinedNan)
    check_lines(([None],), LineType.ChunkCombinedNan)
    check_lines(([None, points],), LineType.ChunkCombinedNan)

    with pytest.raises(TypeError, match=r"Expected tuple not <class 'list'>"):
        check_lines(list(), LineType.ChunkCombinedNan)
    with pytest.raises(ValueError, match=r"Expected tuple of length 1 not 2"):
        check_lines(([], []), LineType.ChunkCombinedNan)
    with pytest.raises(TypeError, match=r"Expected tuple to contain 1 lists"):
        check_lines((None,), LineType.ChunkCombinedNan)
    with pytest.raises(ValueError, match=r"Expected 1 non-empty lists"):
        check_lines(([],), LineType.ChunkCombinedNan)

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        check_lines(([list(points)],), LineType.ChunkCombinedNan)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import numpy as np
from numpy.testing import assert_allclose, assert_array_equal
import pytest

from contourpy import FillType, LineType, contour_generator, dechunk_filled, dechunk_lines

from . import util_test

if TYPE_CHECKING:
    import contourpy._contourpy as cpy


@pytest.fixture
def z() -> cpy.CoordinateArray:
    return np.array([[0, 1, 0, 0, 0],
                     [0, 1, 0, 0, 0],
                     [0, 1, 0, 0, 0],
                     [0, 2, 0, 1, 0],
                     [0, 0, 0, 0, 0]], dtype=np.float64)


@pytest.mark.parametrize("fill_type", FillType.__members__.values())
@pytest.mark.parametrize("chunk_size", (0, 2))
def test_dechunk_filled(z: cpy.CoordinateArray, fill_type: FillType, chunk_size: int) -> None:
    cont_gen = contour_generator(z=z, fill_type=fill_type, chunk_size=chunk_size)
    assert cont_gen.chunk_count == ((2, 2) if chunk_size==2 else (1, 1))
    filled = cont_gen.filled(0.5, 1.5)

    dechunked = dechunk_filled(filled, fill_type)
    util_test.assert_filled(dechunked, fill_type)

    if chunk_size == 0 or fill_type in (FillType.OuterCode, FillType.OuterOffset):
        # Dechunking is a no-op for non-chunked fill types.
        assert dechunked == filled
    else:
        nchunks = len(dechunked[0])
        assert nchunks == 1

        # Note it is important that this contains a polygon with a hole,
        # and using chunk_size=2 there is an empty chunk.
        expected_points = np.array([
            [0.5, 0], [1, 0], [1.5, 0], [1.5, 1], [1.5, 2], [1, 2], [0.5, 2], [0.5, 1], [0.5, 0],
            [0.5, 2], [1, 2], [1.5, 2], [1.75, 3], [1, 3.75], [0.25, 3], [0.5, 2], [1, 2.5],
            [0.75, 3], [1, 3.25], [1.25, 3], [1, 2.5], [2.5, 3], [3, 2.5], [3.5, 3], [3, 3.5],
            [2.5, 3]])
        expected_codes = np.array([1, 2, 2, 2, 2, 2, 2, 2, 79, 1, 2, 2, 2, 2, 2, 79, 1, 2, 2, 2, 79,
                                   1, 2, 2, 2, 79])
        expected_offsets = np.array([0, 9, 16, 21, 26])

        assert dechunked[0][0] is not None
        assert_allclose(dechunked[0][0], expected_points)

        if fill_type == FillType.ChunkCombinedCode:
            if TYPE_CHECKING:
                dechunked = cast(cpy.FillReturn_ChunkCombinedCode, dechunked)
            assert dechunked[1][0] is not None
            assert_array_equal(dechunked[1][0], expected_codes)
        elif fill_type == FillType.ChunkCombinedOffset:
            if TYPE_CHECKING:
                dechunked = cast(cpy.FillReturn_ChunkCombinedOffset, dechunked)
            assert dechunked[1][0] is not None
            assert_array_equal(dechunked[1][0], expected_offsets)
        elif fill_type == FillType.ChunkCombinedCodeOffset:
            if TYPE_CHECKING:
                dechunked = cast(cpy.FillReturn_ChunkCombinedCodeOffset, dechunked)
            assert dechunked[1][0] is not None
            assert_array_equal(dechunked[1][0], expected_codes)
            assert dechunked[2][0] is not None
            assert_array_equal(dechunked[2][0], [0, 9, 21, 26])
        elif fill_type == FillType.ChunkCombinedOffsetOffset:
            if TYPE_CHECKING:
                dechunked = cast(cpy.FillReturn_ChunkCombinedOffsetOffset, dechunked)
            assert dechunked[1][0] is not None
            assert_array_equal(dechunked[1][0], expected_offsets)
            assert dechunked[2][0] is not None
            assert_array_equal(dechunked[2][0], [0, 1, 3, 4])
        else:
            raise RuntimeError(f"Unexpected fill_type {fill_type}")


@pytest.mark.parametrize("fill_type", FillType.__members__.values())
@pytest.mark.parametrize("chunk_size", (0, 1))
def test_dechunk_filled_empty(z: cpy.CoordinateArray, fill_type: FillType, chunk_size: int) -> None:
    cont_gen = contour_generator(z=z, fill_type=fill_type, chunk_size=chunk_size)
    filled = cont_gen.filled(3, 4)

    dechunked = dechunk_filled(filled, fill_type)
    util_test.assert_filled(dechunked, fill_type)

    if fill_type in (FillType.OuterCode, FillType.OuterOffset):
        assert dechunked == ([], [])
    elif fill_type in (FillType.ChunkCombinedCode, FillType.ChunkCombinedOffset):
        assert dechunked == ([None], [None])
    elif fill_type in (FillType.ChunkCombinedCodeOffset, FillType.ChunkCombinedOffsetOffset):
        assert dechunked == ([None], [None], [None])
    else:
        raise RuntimeError(f"Unexpected fill_type {fill_type}")


@pytest.mark.parametrize("line_type", LineType.__members__.values())
@pytest.mark.parametrize("chunk_size", (0, 2))
def test_dechunk_lines(z: cpy.CoordinateArray, line_type: LineType, chunk_size: int) -> None:
    cont_gen = contour_generator(z=z, line_type=line_type, chunk_size=chunk_size)
    assert cont_gen.chunk_count == ((2, 2) if chunk_size==2 else (1, 1))
    lines = cont_gen.lines(0.5)

    dechunked = dechunk_lines(lines, line_type)
    util_test.assert_lines(dechunked, line_type)

    if chunk_size == 0 or line_type in (LineType.Separate, LineType.SeparateCode):
        # Dechunking is a no-op for non-chunked line types.
        assert dechunked == lines
    else:
        nchunks = len(dechunked[0])
        assert nchunks == 1

        # Note it is important that this contains a closed line loop as well as open line strips,
        # and using chunk_size=2 there is an empty chunk.
        expected_points = np.array([
            [1.5, 0], [1.5, 1], [1.5, 2], [0.5, 2], [0.5, 1], [0.5, 0], [1.5, 2], [1.75, 3],
            [1, 3.75], [0.25, 3], [0.5, 2], [2.5, 3], [3, 2.5], [3.5, 3], [3, 3.5], [2.5, 3]])
        expected_codes = np.array([1, 2, 2, 1, 2, 2, 1, 2, 2, 2, 2, 1, 2, 2, 2, 79])
        expected_offsets = np.array([0, 3, 6, 11, 16])

        if line_type == LineType.ChunkCombinedCode:
            if TYPE_CHECKING:
                dechunked = cast(cpy.LineReturn_ChunkCombinedCode, dechunked)
            assert dechunked[0][0] is not None
            assert_allclose(dechunked[0][0], expected_points)
            assert dechunked[1][0] is not None
            assert_array_equal(dechunked[1][0], expected_codes)
        elif line_type == LineType.ChunkCombinedOffset:
            if TYPE_CHECKING:
                dechunked = cast(cpy.LineReturn_ChunkCombinedOffset, dechunked)
            assert dechunked[0][0] is not None
            assert_allclose(dechunked[0][0], expected_points)
            assert dechunked[1][0] is not None
            assert_array_equal(dechunked[1][0], expected_offsets)
        elif line_type == LineType.ChunkCombinedNan:
            if TYPE_CHECKING:
                dechunked = cast(cpy.LineReturn_ChunkCombinedNan, dechunked)
            assert dechunked[0][0] is not None
            # Convert offsets to int64 to avoid numpy error when mixing signed and unsigned ints.
            expected = np.insert(expected_points, expected_offsets[1:-1].astype(np.int64),
                                 [np.nan, np.nan], axis=0)
            assert_allclose(dechunked[0][0], expected)
        else:
            raise RuntimeError(f"Unexpected line_type {line_type}")


@pytest.mark.parametrize("line_type", LineType.__members__.values())
@pytest.mark.parametrize("chunk_size", (0, 1))
def test_dechunk_lines_empty(z: cpy.CoordinateArray, line_type: LineType, chunk_size: int) -> None:
    cont_gen = contour_generator(z=z, line_type=line_type, chunk_size=chunk_size)
    lines = cont_gen.lines(4)

    dechunked = dechunk_lines(lines, line_type)
    util_test.assert_lines(dechunked, line_type)

    if line_type == LineType.Separate:
        assert dechunked == []
    elif line_type == LineType.SeparateCode:
        assert dechunked == ([], [])
    elif line_type in (LineType.ChunkCombinedCode, LineType.ChunkCombinedOffset):
        assert dechunked == ([None], [None])
    elif line_type == LineType.ChunkCombinedNan:
        assert dechunked == ([None],)
    else:
        raise RuntimeError(f"Unexpected line_type {line_type}")



from __future__ import annotations

import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal
import pytest

import contourpy.array as arr
from contourpy.types import code_dtype, offset_dtype, point_dtype

from . import util_test


def test_codes_from_offsets() -> None:
    codes = arr.codes_from_offsets(np.array([0, 3], dtype=offset_dtype))
    util_test.assert_code_array(codes, 3)
    assert_array_equal(codes, [1, 2, 79])

    codes = arr.codes_from_offsets(np.array([0, 3, 7], dtype=offset_dtype))
    util_test.assert_code_array(codes, 7)
    assert_array_equal(codes, [1, 2, 79, 1, 2, 2, 79])

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        arr.codes_from_offsets([0, 3, 7])  # type: ignore[arg-type]

    with pytest.raises(ValueError, match=r"Expected numpy array of dtype <class 'numpy.uint32'>"):
        arr.codes_from_offsets(np.array([0, 3, 7], dtype=np.int64))

    with pytest.raises(ValueError, match=r"Expected numpy array of shape"):
        arr.codes_from_offsets(np.array([[0, 3], [7, 10]], dtype=offset_dtype))

    with pytest.raises(ValueError, match=r"First element of offset array must be 0"):
        arr.codes_from_offsets(np.array([1, 2], dtype=offset_dtype))


def test_codes_from_offsets_and_points() -> None:
    codes = arr.codes_from_offsets_and_points(
        np.array([0, 3], dtype=offset_dtype),
        np.array([[0, 1], [2, 3], [4, 5]], dtype=point_dtype),
    )
    util_test.assert_code_array(codes, 3)
    assert_array_equal(codes, [1, 2, 2])

    codes = arr.codes_from_offsets_and_points(
        np.array([0, 3], dtype=offset_dtype),
        np.array([[0, 1], [2, 3], [0, 1]], dtype=point_dtype),
    )
    util_test.assert_code_array(codes, 3)
    assert_array_equal(codes, [1, 2, 79])

    codes = arr.codes_from_offsets_and_points(
        np.array([0, 3, 7], dtype=offset_dtype),
        np.array([[0, 1], [2, 3], [0, 1], [5, 6], [7, 8], [9, 10], [5, 6]], dtype=point_dtype),
    )
    util_test.assert_code_array(codes, 7)
    assert_array_equal(codes, [1, 2, 79, 1, 2, 2, 79])

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        arr.codes_from_offsets_and_points(
            [0, 2],  # type: ignore[arg-type]
            np.array([[0, 1], [2, 3]], dtype=point_dtype),
        )

    with pytest.raises(ValueError, match=r"Expected numpy array of dtype <class 'numpy.uint32'>"):
        arr.codes_from_offsets_and_points(
            np.array([0, 2], dtype=np.int64),
            np.array([[0, 1], [2, 3]], dtype=point_dtype),
        )

    with pytest.raises(ValueError, match=r"Expected numpy array of shape"):
        arr.codes_from_offsets_and_points(
            np.array([[0], [2]], dtype=offset_dtype),
            np.array([[0, 1], [2, 3]], dtype=point_dtype),
        )

    with pytest.raises(ValueError, match=r"First element of offset array must be 0"):
        arr.codes_from_offsets_and_points(
            np.array([1, 2], dtype=offset_dtype),
            np.array([[0, 1], [2, 3]], dtype=point_dtype),
        )

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        arr.codes_from_offsets_and_points(
            np.array([0, 2],
            dtype=offset_dtype), [[0, 1], [2, 3]],  # type: ignore[arg-type]
        )

    with pytest.raises(ValueError, match=r"Expected numpy array of dtype <class 'numpy.float64'>"):
        arr.codes_from_offsets_and_points(
            np.array([0, 2], dtype=offset_dtype),
            np.array([[0, 1], [2, 3]], dtype=np.float32),
        )

    with pytest.raises(ValueError, match=r"Expected numpy array of shape"):
        arr.codes_from_offsets_and_points(
            np.array([0, 2], dtype=offset_dtype),
            np.array([0, 1, 2, 3], dtype=point_dtype),
        )


def test_codes_from_points() -> None:
    codes = arr.codes_from_points(np.array([[0, 1], [2, 3], [4, 5]], dtype=point_dtype))
    util_test.assert_code_array(codes, 3)
    assert_array_equal(codes, [1, 2, 2])

    codes = arr.codes_from_points(np.array([[0, 1], [2, 3], [0, 1]], dtype=point_dtype))
    util_test.assert_code_array(codes, 3)
    assert_array_equal(codes, [1, 2, 79])

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        arr.codes_from_points([[0, 1], [2, 3], [0, 1]])  # type: ignore[arg-type]

    with pytest.raises(ValueError, match=r"Expected numpy array of dtype <class 'numpy.float64'>"):
        arr.codes_from_points(np.array([[0, 1], [2, 3], [4, 5]], dtype=np.float32))

    with pytest.raises(ValueError, match=r"Expected numpy array of shape"):
        arr.codes_from_points(np.array([0, 1, 2, 3], dtype=point_dtype))


def test_concat_codes() -> None:
    codes0 = np.array([1, 2, 79], dtype=code_dtype)
    codes1 = np.array([1, 2, 2, 2], dtype=code_dtype)

    codes = arr.concat_codes([codes0])
    util_test.assert_code_array(codes, 3)
    assert_array_equal(codes, codes0)

    codes = arr.concat_codes([codes1])
    util_test.assert_code_array(codes, 4)
    assert_array_equal(codes, codes1)

    codes = arr.concat_codes([codes0, codes1])
    util_test.assert_code_array(codes, 7)
    assert_array_equal(codes, [1, 2, 79, 1, 2, 2, 2])

    codes = arr.concat_codes([codes1, codes0])
    util_test.assert_code_array(codes, 7)
    assert_array_equal(codes, [1, 2, 2, 2, 1, 2, 79])

    with pytest.raises(ValueError, match=r"Empty list passed to concat_codes"):
        arr.concat_codes([])


def test_concat_codes_or_none() -> None:
    codes0 = np.array([1, 2, 79], dtype=code_dtype)
    codes1 = np.array([1, 2, 2, 2], dtype=code_dtype)

    assert arr.concat_codes_or_none([None]) is None
    assert arr.concat_codes_or_none([None, None]) is None
    assert arr.concat_codes_or_none([None, None, None]) is None

    codes = arr.concat_codes_or_none([None, codes0, None])
    assert codes is not None
    util_test.assert_code_array(codes, 3)
    assert_array_equal(codes, [1, 2, 79])

    codes = arr.concat_codes_or_none([None, codes0, None, codes1, None])
    assert codes is not None
    util_test.assert_code_array(codes, 7)
    assert_array_equal(codes, [1, 2, 79, 1, 2, 2, 2])

    codes = arr.concat_codes_or_none([None, codes1, None, codes0, None])
    assert codes is not None
    util_test.assert_code_array(codes, 7)
    assert_array_equal(codes, [1, 2, 2, 2, 1, 2, 79])


def test_concat_offsets() -> None:
    offsets0 = np.array([0, 2, 5], dtype=offset_dtype)
    offsets1 = np.array([0, 3, 7], dtype=offset_dtype)
    offsets2 = np.array([0, 4, 8], dtype=offset_dtype)

    offsets = arr.concat_offsets([offsets0, offsets1])
    util_test.assert_offset_array(offsets, 12)
    assert_array_equal(offsets, [0, 2, 5, 8, 12])

    offsets = arr.concat_offsets([offsets0, offsets1, offsets2])
    util_test.assert_offset_array(offsets, 20)
    assert_array_equal(offsets, [0, 2, 5, 8, 12, 16, 20])

    with pytest.raises(ValueError, match=r"Empty list passed to concat_offsets"):
        arr.concat_offsets([])


def test_concat_offsets_or_none() -> None:
    offsets0 = np.array([0, 2, 5], dtype=offset_dtype)
    offsets1 = np.array([0, 3, 7], dtype=offset_dtype)
    offsets2 = np.array([0, 4, 8], dtype=offset_dtype)

    assert arr.concat_offsets_or_none([None]) is None
    assert arr.concat_offsets_or_none([None, None]) is None
    assert arr.concat_offsets_or_none([None, None, None]) is None

    offsets = arr.concat_offsets_or_none([None, offsets0, None, offsets1, None])
    assert offsets is not None
    util_test.assert_offset_array(offsets, 12)
    assert_array_equal(offsets, [0, 2, 5, 8, 12])

    offsets = arr.concat_offsets_or_none([None, None, offsets0, offsets1, None, offsets2])
    assert offsets is not None
    util_test.assert_offset_array(offsets, 20)
    assert_array_equal(offsets, [0, 2, 5, 8, 12, 16, 20])


def test_concat_points() -> None:
    points0 = np.array([[0, 1], [2, 3], [4, 5]], dtype=point_dtype)
    points1 = np.array([[6, 7], [8, 9]], dtype=point_dtype)

    points = arr.concat_points([points0])
    util_test.assert_point_array(points)
    assert_array_almost_equal(points, points0)

    points = arr.concat_points([points0, points1])
    util_test.assert_point_array(points)
    assert_array_almost_equal(
        points,
        np.array([[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]], dtype=point_dtype),
    )

    points = arr.concat_points([points1, points0])
    util_test.assert_point_array(points)
    assert_array_almost_equal(points, [[6, 7], [8, 9], [0, 1], [2, 3], [4, 5]])

    with pytest.raises(ValueError, match=r"Empty list passed to concat_points"):
        arr.concat_points([])


def test_concat_points_or_none() -> None:
    points0 = np.array([[0, 1], [2, 3], [4, 5]], dtype=point_dtype)
    points1 = np.array([[6, 7], [8, 9]], dtype=point_dtype)

    assert arr.concat_points_or_none([None]) is None
    assert arr.concat_points_or_none([None, None]) is None
    assert arr.concat_points_or_none([None, None, None]) is None

    points = arr.concat_points_or_none([None, points0, None])
    assert points is not None
    util_test.assert_point_array(points)
    assert_array_almost_equal(points, points0)

    points = arr.concat_points_or_none([None, points0, None, points1, None])
    assert points is not None
    util_test.assert_point_array(points)
    assert_array_almost_equal(points, [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]])

    points = arr.concat_points_or_none([None, None, points1, points0, None])
    assert points is not None
    util_test.assert_point_array(points)
    assert_array_almost_equal(points, [[6, 7], [8, 9], [0, 1], [2, 3], [4, 5]])


def test_concat_points_or_none_with_nan() -> None:
    points0 = np.array([[0, 1], [2, 3], [4, 5]], dtype=point_dtype)
    points1 = np.array([[6, 7], [8, 9]], dtype=point_dtype)
    points2 = np.array([[-1, -2], [-3, -4]], dtype=point_dtype)

    assert arr.concat_points_or_none_with_nan([None]) is None
    assert arr.concat_points_or_none_with_nan([None, None]) is None
    assert arr.concat_points_or_none_with_nan([None, None, None]) is None

    points = arr.concat_points_or_none_with_nan([None, points0, None])
    assert points is not None
    util_test.assert_point_array(points)
    assert_array_almost_equal(points, points0)

    points = arr.concat_points_or_none_with_nan([None, points0, None, None, points1, None])
    assert points is not None
    util_test.assert_point_array(points, allow_nan=True)
    assert_array_almost_equal(
        points,
        np.array([[0, 1], [2, 3], [4, 5], [np.nan, np.nan], [6, 7], [8, 9]], dtype=point_dtype),
    )

    points = arr.concat_points_or_none_with_nan([None, points0, None, points1, None, None, points2])
    assert points is not None
    util_test.assert_point_array(points, allow_nan=True)
    assert_array_almost_equal(
        points,
        np.array([[0, 1], [2, 3], [4, 5], [np.nan, np.nan], [6, 7], [8, 9], [np.nan, np.nan],
                  [-1, -2], [-3, -4]], dtype=point_dtype),
    )


def test_concat_points_with_nan() -> None:
    points0 = np.array([[0, 1], [2, 3], [4, 5]], dtype=point_dtype)
    points1 = np.array([[6, 7], [8, 9]], dtype=point_dtype)
    points2 = np.array([[-1, -2], [-3, -4]], dtype=point_dtype)

    points = arr.concat_points_with_nan([points0])
    util_test.assert_point_array(points)
    assert_array_almost_equal(points, points0)

    points = arr.concat_points_with_nan([points0, points1])
    util_test.assert_point_array(points, allow_nan=True)
    assert_array_almost_equal(
        points,
        np.array([[0, 1], [2, 3], [4, 5], [np.nan, np.nan], [6, 7], [8, 9]], dtype=point_dtype),
    )

    points = arr.concat_points_with_nan([points0, points1, points2])
    util_test.assert_point_array(points, allow_nan=True)
    assert_array_almost_equal(
        points,
        np.array([[0, 1], [2, 3], [4, 5], [np.nan, np.nan], [6, 7], [8, 9], [np.nan, np.nan],
                  [-1, -2], [-3, -4]], dtype=point_dtype),
    )

    with pytest.raises(ValueError, match=r"Empty list passed to concat_points_with_nan"):
        arr.concat_points_with_nan([])


def test_insert_nan_at_offsets() -> None:
    points0 = np.array([[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]], dtype=point_dtype)

    points = arr.insert_nan_at_offsets(points0, np.array([0, 5], dtype=offset_dtype))
    util_test.assert_point_array(points, allow_nan=False)
    assert_array_almost_equal(points, points0)

    points = arr.insert_nan_at_offsets(points0, np.array([0, 1, 5], dtype=offset_dtype))
    util_test.assert_point_array(points, allow_nan=True)
    assert_array_almost_equal(
        points,
        np.array([[0, 1], [np.nan, np.nan], [2, 3], [4, 5], [6, 7], [8, 9]], dtype=point_dtype),
    )

    points = arr.insert_nan_at_offsets(points0, np.array([0, 1, 4, 5], dtype=offset_dtype))
    util_test.assert_point_array(points, allow_nan=True)
    assert_array_almost_equal(
        points,
        np.array([[0, 1], [np.nan, np.nan], [2, 3], [4, 5], [6, 7], [np.nan, np.nan], [8, 9]],
                 dtype=point_dtype),
    )

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        arr.insert_nan_at_offsets(
            [[0, 1], [2, 3], [4, 5]],  # type: ignore[arg-type]
            np.array([0, 3], dtype=offset_dtype),
        )

    with pytest.raises(ValueError, match=r"Expected numpy array of dtype <class 'numpy.float64'>"):
        arr.insert_nan_at_offsets(
            np.array([[0, 1], [2, 3], [4, 5]], dtype=np.float32),
            np.array([0, 3], dtype=offset_dtype),
        )

    with pytest.raises(ValueError, match=r"Expected numpy array of shape"):
        arr.insert_nan_at_offsets(
            np.array([0, 1, 2, 3, 4, 5], dtype=point_dtype),
            np.array([0, 3], dtype=offset_dtype),
        )

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        arr.insert_nan_at_offsets(
            np.array([[0, 1], [2, 3], [4, 5]], dtype=point_dtype),
            [0, 3],  # type: ignore[arg-type]
        )

    with pytest.raises(ValueError, match=r"Expected numpy array of dtype <class 'numpy.uint32'>"):
        arr.insert_nan_at_offsets(
            np.array([[0, 1], [2, 3], [4, 5]], dtype=point_dtype),
            np.array([0, 3], dtype=np.int64),
        )

    with pytest.raises(ValueError, match=r"Expected numpy array of shape"):
        arr.insert_nan_at_offsets(
            np.array([[0, 1], [2, 3], [4, 5]], dtype=point_dtype),
            np.array([[0, 3]], dtype=offset_dtype),
        )

    with pytest.raises(ValueError, match=r"First element of offset array must be 0"):
        arr.insert_nan_at_offsets(
            np.array([[0, 1], [2, 3], [4, 5]], dtype=point_dtype),
            np.array([1, 3], dtype=offset_dtype),
        )


def test_offsets_from_codes() -> None:
    offsets = arr.offsets_from_codes(np.array([1, 2, 79], dtype=code_dtype))
    util_test.assert_offset_array(offsets, 3)
    assert_array_equal(offsets, [0, 3])

    offsets = arr.offsets_from_codes(np.array([1, 2, 79, 1, 2, 2, 1, 2], dtype=code_dtype))
    util_test.assert_offset_array(offsets, 8)
    assert_array_equal(offsets, [0, 3, 6, 8])

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        arr.offsets_from_codes([1, 2, 79])  # type: ignore[arg-type]

    with pytest.raises(ValueError, match=r"Expected numpy array of dtype <class 'numpy.uint8'>"):
        arr.offsets_from_codes(np.array([1, 2, 79], dtype=np.int64))

    with pytest.raises(ValueError, match=r"Expected numpy array of shape"):
        arr.offsets_from_codes(np.array([[1, 2], [2, 79]], dtype=code_dtype))

    with pytest.raises(ValueError, match=r"First element of code array must be 1"):
        arr.offsets_from_codes(np.array([2, 79], dtype=code_dtype))


def test_offsets_from_lengths() -> None:
    points0 = np.array([[0, 1], [2, 3], [4, 5]], dtype=point_dtype)
    points1 = np.array([[6, 7], [8, 9]], dtype=point_dtype)

    offsets = arr.offsets_from_lengths([points0])
    util_test.assert_offset_array(offsets, 3)
    assert_array_equal(offsets, [0, 3])

    offsets = arr.offsets_from_lengths([points0, points1])
    util_test.assert_offset_array(offsets, 5)
    assert_array_equal(offsets, [0, 3, 5])

    offsets = arr.offsets_from_lengths([points1, points0])
    util_test.assert_offset_array(offsets, 5)
    assert_array_equal(offsets, [0, 2, 5])

    with pytest.raises(ValueError, match=r"Empty list passed to offsets_from_lengths"):
        arr.offsets_from_lengths([])


def test_outer_offsets_from_list_of_codes() -> None:
    codes0 = np.array([1, 2, 79, 1, 2], dtype=code_dtype)
    codes1 = np.array([1, 2, 2, 2, 1, 2, 1, 2, 2], dtype=code_dtype)

    outer_offsets = arr.outer_offsets_from_list_of_codes([codes0])
    util_test.assert_offset_array(outer_offsets, 2)
    assert_array_equal(outer_offsets, [0, 2])

    outer_offsets = arr.outer_offsets_from_list_of_codes([codes1])
    util_test.assert_offset_array(outer_offsets, 3)
    assert_array_equal(outer_offsets, [0, 3])

    outer_offsets = arr.outer_offsets_from_list_of_codes([codes0, codes1])
    util_test.assert_offset_array(outer_offsets, 5)
    assert_array_equal(outer_offsets, [0, 2, 5])

    outer_offsets = arr.outer_offsets_from_list_of_codes([codes1, codes0])
    util_test.assert_offset_array(outer_offsets, 5)
    assert_array_equal(outer_offsets, [0, 3, 5])

    with pytest.raises(ValueError, match=r"Empty list passed to outer_offsets_from_list_of_codes"):
        arr.outer_offsets_from_list_of_codes([])


def test_outer_offsets_from_list_of_offsets() -> None:
    offsets0 = np.array([0, 3, 5], dtype=offset_dtype)
    offsets1 = np.array([0, 4, 9, 11], dtype=offset_dtype)

    outer_offsets = arr.outer_offsets_from_list_of_offsets([offsets0])
    util_test.assert_offset_array(outer_offsets, 2)
    assert_array_equal(outer_offsets, [0, 2])

    outer_offsets = arr.outer_offsets_from_list_of_offsets([offsets1])
    util_test.assert_offset_array(outer_offsets, 3)
    assert_array_equal(outer_offsets, [0, 3])

    outer_offsets = arr.outer_offsets_from_list_of_offsets([offsets0, offsets1])
    util_test.assert_offset_array(outer_offsets, 5)
    assert_array_equal(outer_offsets, [0, 2, 5])

    outer_offsets = arr.outer_offsets_from_list_of_offsets([offsets1, offsets0])
    util_test.assert_offset_array(outer_offsets, 5)
    assert_array_equal(outer_offsets, [0, 3, 5])

    with pytest.raises(ValueError,
                       match=r"Empty list passed to outer_offsets_from_list_of_offsets"):
        arr.outer_offsets_from_list_of_offsets([])


def test_remove_nan() -> None:
    points0 = np.array([[1.1, 2.2], [3.3, 4.4], [1.1, 2.2]], dtype=point_dtype)
    points1 = np.array([[0, 1], [2, 3], [4, 5], [np.nan, np.nan], [6, 7], [8, 9], [np.nan, np.nan],
                        [-1, -2], [-3, -4], [-5, -6]], dtype=point_dtype)

    points, offsets = arr.remove_nan(points0)
    util_test.assert_point_array(points, allow_nan=False)
    util_test.assert_offset_array(offsets, 3)
    assert_array_almost_equal(points, points0)
    assert_array_equal(offsets, [0, 3])

    points, offsets = arr.remove_nan(points1)
    util_test.assert_point_array(points, allow_nan=True)
    util_test.assert_offset_array(offsets, 8)
    assert_array_almost_equal(
        points, [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9], [-1, -2], [-3, -4], [-5, -6]],
    )
    assert_array_equal(offsets, [0, 3, 5, 8])


def test_split_codes_by_offsets() -> None:
    codes = np.array([1, 2, 1, 2, 79], dtype=code_dtype)

    list_of_codes = arr.split_codes_by_offsets(codes, np.array([0, 5], dtype=offset_dtype))
    assert isinstance(list_of_codes, list)
    assert len(list_of_codes) == 1
    util_test.assert_code_array(list_of_codes[0], 5)
    assert_array_equal(list_of_codes[0], codes)

    list_of_codes = arr.split_codes_by_offsets(codes, np.array([0, 2, 5], dtype=offset_dtype))
    assert isinstance(list_of_codes, list)
    assert len(list_of_codes) == 2
    util_test.assert_code_array(list_of_codes[0], 2)
    util_test.assert_code_array(list_of_codes[1], 3)
    assert_array_equal(list_of_codes[0], [1, 2])
    assert_array_equal(list_of_codes[1], [1, 2, 79])

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
       arr.split_codes_by_offsets(codes, [0, 5])  # type: ignore[arg-type]

    with pytest.raises(ValueError, match=r"Expected numpy array of dtype <class 'numpy.uint32'>"):
        arr.codes_from_offsets(np.array([0, 5], dtype=np.int64))

    with pytest.raises(ValueError, match=r"Expected numpy array of shape"):
        arr.codes_from_offsets(np.array([[0, 5]], dtype=offset_dtype))

    with pytest.raises(ValueError, match=r"First element of offset array must be 0"):
        arr.codes_from_offsets(np.array([1, 2], dtype=offset_dtype))


def test_split_points_at_nan() -> None:
    points0 = np.array([[1.1, 2.2], [3.3, 4.4]], dtype=point_dtype)
    points1 = np.array([[0, 1], [2, 3], [np.nan, np.nan], [4, 5], [6, 7], [8, 9], [np.nan, np.nan],
                        [10, 11], [12, 13]], dtype=point_dtype)

    list_of_points = arr.split_points_at_nan(points0)
    assert isinstance(list_of_points, list)
    assert len(list_of_points) == 1
    util_test.assert_point_array(list_of_points[0])
    assert_array_almost_equal(list_of_points[0], points0)

    list_of_points = arr.split_points_at_nan(points1)
    assert isinstance(list_of_points, list)
    assert len(list_of_points) == 3
    for points in list_of_points:
        util_test.assert_point_array(points)
    assert_array_almost_equal(list_of_points[0], [[0, 1], [2, 3]])
    assert_array_almost_equal(list_of_points[1], [[4, 5], [6, 7], [8, 9]])
    assert_array_almost_equal(list_of_points[2], [[10, 11], [12, 13]])

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
        arr.split_points_at_nan([[0, 1], [2, 3]])  # type: ignore[arg-type]

    with pytest.raises(ValueError, match=r"Expected numpy array of dtype <class 'numpy.float64'>"):
        arr.split_points_at_nan(np.array([[0, 1], [2, 3]], dtype=np.float32))

    with pytest.raises(ValueError, match=r"Expected numpy array of shape"):
        arr.split_points_at_nan(np.array([[0, 1, 2, 3]], dtype=point_dtype))


def test_split_points_by_offsets() -> None:
    points = np.array([[0, 1], [2, 3], [4, 5], [6, 7], [8, 9], [10, 11]], dtype=point_dtype)

    list_of_points = arr.split_points_by_offsets(points, np.array([0, 6], dtype=offset_dtype))
    assert isinstance(list_of_points, list)
    assert len(list_of_points) == 1
    util_test.assert_point_array(list_of_points[0])
    assert_array_almost_equal(list_of_points[0], points)

    list_of_points = arr.split_points_by_offsets(points, np.array([0, 2, 6], dtype=offset_dtype))
    assert isinstance(list_of_points, list)
    assert len(list_of_points) == 2
    util_test.assert_point_array(list_of_points[0])
    util_test.assert_point_array(list_of_points[1])
    assert_array_almost_equal(list_of_points[0], [[0, 1], [2, 3]])
    assert_array_almost_equal(list_of_points[1], [[4, 5], [6, 7], [8, 9], [10, 11]])

    with pytest.raises(TypeError, match=r"Expected numpy array not <class 'list'>"):
       arr.split_points_by_offsets(points, [0, 6])  # type: ignore[arg-type]

    with pytest.raises(ValueError, match=r"Expected numpy array of dtype <class 'numpy.uint32'>"):
       arr.split_points_by_offsets(points, np.array([0, 6], dtype=np.int64))

    with pytest.raises(ValueError, match=r"Expected numpy array of shape"):
       arr.split_points_by_offsets(points, np.array([[0, 6]], dtype=offset_dtype))

    with pytest.raises(ValueError, match=r"First element of offset array must be 0"):
       arr.split_points_by_offsets(points, np.array([1, 6], dtype=offset_dtype))



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pytest

from contourpy import (
    FillType, LineType, contour_generator, convert_filled, convert_lines, dechunk_filled,
    dechunk_lines,
)

from . import util_test

if TYPE_CHECKING:
    import contourpy._contourpy as cpy


@pytest.fixture
def z() -> cpy.CoordinateArray:
    # Care needed with test data as although arbitrary z produces identical results for lines
    # regardless of line_type, this is not the case for filled as the order that polygons are
    # produced is different depending on whether the relationship between outer and inner
    # boundaries is calculated. If it is then inner boundaries directly follow their outer boundary,
    # if not then the boundaries are returned in the order they are found.
    # This test data is chosen so that filled results are always the same.
    return np.array([
        [1, 0, 0, 0, 0, 0, 0],
        [0, 0, 2, 0, 0, 0, 0],
        [0, 2, 0, 0, 0, 0, 0],
        [0, 1, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 1],
        [0, 2, 2, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 1],
    ], dtype=np.float64)


@pytest.mark.parametrize("fill_type_to", FillType.__members__.values())
@pytest.mark.parametrize("fill_type_from", FillType.__members__.values())
@pytest.mark.parametrize("chunk_size", (0, 2, 3))
@pytest.mark.parametrize("empty", (False, True))
def test_convert_filled(
    z: cpy.CoordinateArray,
    fill_type_from: FillType,
    fill_type_to: FillType,
    chunk_size: int,
    empty: bool,
) -> None:
    if empty:
        # Levels outside of z bounds so no polygons produced.
        lower_level, upper_level = 3.0, 4.0
    else:
        # Using chunk_size=2 or 3 there is an empty chunk.
        lower_level, upper_level = 0.5, 1.5

    filled = contour_generator(z=z, fill_type=fill_type_from, chunk_size=chunk_size).filled(
        lower_level, upper_level)

    if (fill_type_from in (FillType.ChunkCombinedCode, FillType.ChunkCombinedOffset) and
        fill_type_to not in (FillType.ChunkCombinedCode, FillType.ChunkCombinedOffset)):
        msg = f"Conversion from {fill_type_from} to {fill_type_to} not supported"
        with pytest.raises(ValueError, match=msg):
            convert_filled(filled, fill_type_from, fill_type_to)
    else:
        converted = convert_filled(filled, fill_type_from, fill_type_to)
        util_test.assert_filled(converted, fill_type_to)

        compare = contour_generator(z=z, fill_type=fill_type_to, chunk_size=chunk_size).filled(
            lower_level, upper_level)

        # Converting OuterCode or OuterOffset to any ChunkCombined* returns all polygons in a single
        # chunk, so need to compare to dechunked_filled(compare).
        if (fill_type_from in (FillType.OuterCode, FillType.OuterOffset) and
            fill_type_to not in (FillType.OuterCode, FillType.OuterOffset)):
            compare = dechunk_filled(compare, fill_type_to)

            util_test.assert_equal_recursive(converted, compare)


@pytest.mark.parametrize("line_type_to", LineType.__members__.values())
@pytest.mark.parametrize("line_type_from", LineType.__members__.values())
@pytest.mark.parametrize("chunk_size", (0, 2, 3))
@pytest.mark.parametrize("empty", (False, True))
def test_convert_lines(
    z: cpy.CoordinateArray,
    line_type_from: LineType,
    line_type_to: LineType,
    chunk_size: int,
    empty: bool,
) -> None:
    if empty:
        # Level outside of z bounds so no lines produced.
        level = 4.0
    else:
        # Lines includes a closed line loop as well as open line strips,
        # and using chunk_size=2 or 3 there is an empty chunk.
        level = 0.5
    lines = contour_generator(z=z, line_type=line_type_from, chunk_size=chunk_size).lines(level)
    converted = convert_lines(lines, line_type_from, line_type_to)

    compare = contour_generator(z=z, line_type=line_type_to, chunk_size=chunk_size).lines(level)
    util_test.assert_lines(converted, line_type_to)

    # Converting Separate or SeparateCode to any ChunkCombined* returns all lines in a single chunk,
    # so need to compare to dechunked_lines(compare).
    if (line_type_from in (LineType.Separate, LineType.SeparateCode) and
        line_type_to not in (LineType.Separate, LineType.SeparateCode)):
       compare = dechunk_lines(compare, line_type_to)

    util_test.assert_equal_recursive(converted, compare)
